/*
**
** File: scc.c -- software implementation of the Konami SCC
** emulator. More than one SCC supported, and SCC+ as well as
** normal megarom SCC.
**
** For information on the scc, see:
**    http://www.msxnet.org/tech/scc.html
** and
**    http://www.msxnet.org/tech/SoundCartridge.html
**
** Todo:
** Check the sound emulation itself (scc-u.c)
**
** By Sean Young <sean@msxnet.org>
*/

#include "scc.h"

#define DATATYPE u16

//#define DATATYPE UINT8
#define DATACONV(A) ( (A) * 3)

/* the struct */
typedef struct KONSCC {
    u8	Regs[17] ;
    u8 	Waves[5*32] ;

    double	UpdateStep;
    int		SampleRate;
    int		Channel;

    /* state variables */
    int         Counter0, Counter1, Counter2, Counter3, Counter4 ;
};

static struct KONSCC SCC[MAX_SCC];

/* the registers */
#define SCC_1FINE       (0)
#define SCC_1COARSE     (1)
#define SCC_2FINE       (2)
#define SCC_2COARSE     (3)
#define SCC_3FINE       (4)
#define SCC_3COARSE     (5)
#define SCC_4FINE       (6)
#define SCC_4COARSE     (7)
#define SCC_5FINE       (8)
#define SCC_5COARSE     (9)
#define SCC_1VOL        (10)
#define SCC_2VOL        (11)
#define SCC_3VOL        (12)
#define SCC_4VOL        (13)
#define SCC_5VOL        (14)
#define SCC_ENABLE      (15)
#define SCC_DEFORM      (16)

/*
** reset all chip registers.
*/
void SCCResetChip(int num)
{
    int i;

    /* initialize hardware registers */
    memset (SCC[num].Regs, 0, 17) ;
    for( i=10; i<15; i++ ) SCC[num].Regs[i] = 10;

    SCC[num].Counter0 = SCC[num].Counter1 = SCC[num].Counter2 = 0;
    SCC[num].Counter3 = SCC[num].Counter4 = 0;

    /* waves memory is not reset */
}

void SCCWriteReg (int n, int r, int v, int t)
{
    /*char buf[40]; */
    static const int Mask[16] = { 0xff, 0xf, 0xff, 0xf, 0xff, 0xf,
	   0xff, 0xf, 0xff, 0xf, 0xf, 0xf, 0xf, 0xf, 0xf, 0x1f } ;
    r &= 0xff;
    v &= 0xff;

    switch (t) {
	case SCC_MEGAROM:
	case SCC_PLUSCOMP:
	    if (r < 0x80) {
		/* 0 - 0x7f is wave forms, 0x60-0x7f for channel 4 and 5 */
		SCC[n].Waves[r] = v ;
		if (r >= 0x60) {
		    SCC[n].Waves[r + 0x20] = v ;
		}
	    } else if (r < 0xa0) {
		/* 9880-988f = 9890-989f */
		SCC[n].Regs[r & 15] = (v & Mask[r & 15]) ;
	    } else switch (t) {
		case SCC_MEGAROM:
		    if (r >= 0xe0) {
			SCC[n].Regs[SCC_DEFORM] = v ;
			//if (v && errorlog) fprintf (errorlog,
			    //"SCC: %02xh written to unemulated register\n", v);

		    }
		    break ;
		case SCC_PLUSCOMP:
		    if ( (r < 0xe0) && (r >= 0xc0) ) {
			SCC[n].Regs[SCC_DEFORM] = v ;
			//if (v && errorlog) fprintf (errorlog,
			  //  "SCC: %02xh written to unemulated register\n", v);
		    }
		    break ;
		}
	    break ;
	case SCC_PLUSEXT:
	    if (r < 0xa0) {
		SCC[n].Waves[r] = v ;
	    } else if (r < 0xc0) {
		SCC[n].Regs[r & 15] = (v & Mask[r & 15]) ;
	    } else if (r < 0xe0) {
		SCC[n].Regs[SCC_DEFORM] = v ;
		//if (v && errorlog) fprintf (errorlog,
		    //"SCC: %02xh written to unemulated register\n", v);
	    }
	    break ;
    }
}

void SCCUpdate16 (int n, INT16 *buffer, int length) {
//#define DATATYPE UINT16
//#define DATACONV(A) ( (A) * 3)
//#include "scc-u.c"
//#undef DATACONV
//#undef DATATYPE
}

void SCCUpdate8 (int n, INT16 *buffer, int length) {
    int	x/*, i*/, c0, c1, l ;
    int	Incr0, Incr1, Incr2, Incr3, Incr4, Vol0, Vol1, Vol2, Vol3, Vol4 ;
#ifdef SCC_INTERPOLATE
    int	c2 ;
#endif
    DATATYPE *buf;

    buf = (DATATYPE*)buffer;

    x = ((int)SCC[n].Regs[SCC_1FINE] + (((int)SCC[n].Regs[SCC_1COARSE])<<8)) ;
    Incr0 = x ? (int)(SCC[n].UpdateStep / (double)x) : 0 ;

    x = ((int)SCC[n].Regs[SCC_2FINE] + (((int)SCC[n].Regs[SCC_2COARSE])<<8)) ;
    Incr1 = x ? (int)(SCC[n].UpdateStep / (double)x) : 0 ;

    x = ((int)SCC[n].Regs[SCC_3FINE] + (((int)SCC[n].Regs[SCC_3COARSE])<<8)) ;
    Incr2 = x ? (int)(SCC[n].UpdateStep / (double)x) : 0 ;

    x = ((int)SCC[n].Regs[SCC_4FINE] + (((int)SCC[n].Regs[SCC_4COARSE])<<8)) ;
    Incr3 = x ? (int)(SCC[n].UpdateStep / (double)x) : 0 ;

    x = ((int)SCC[n].Regs[SCC_5FINE] + (((int)SCC[n].Regs[SCC_5COARSE])<<8)) ;
    Incr4 = x ? (int)(SCC[n].UpdateStep / (double)x) : 0 ;

    Vol0 = (SCC[n].Regs[SCC_ENABLE] & 0x01) ? SCC[n].Regs[SCC_1VOL] : 0 ;
    Vol1 = (SCC[n].Regs[SCC_ENABLE] & 0x02) ? SCC[n].Regs[SCC_2VOL] : 0 ;
    Vol2 = (SCC[n].Regs[SCC_ENABLE] & 0x04) ? SCC[n].Regs[SCC_3VOL] : 0 ;
    Vol3 = (SCC[n].Regs[SCC_ENABLE] & 0x08) ? SCC[n].Regs[SCC_4VOL] : 0 ;
    Vol4 = (SCC[n].Regs[SCC_ENABLE] & 0x10) ? SCC[n].Regs[SCC_5VOL] : 0 ;

    if ( !(Vol0 || Vol1 || Vol2 || Vol3 || Vol4) ) {
        while (length--) *buf++ = DATACONV (0);
	return ;
    }

    while (length--)
    {
	l = 0;
	if (Vol0 && Incr0) {
	    c0 = (SCC[n].Counter0 + Incr0) & 0xffff ;
	    c1 = (signed char)SCC[n].Waves[0x00 + (c0>>11)] ;
#ifdef SCC_INTERPOLATE
	    c2 = (signed char)SCC[n].Waves[0x00 + ((c0+0x800)>>11)] ;
	    l += (Vol0*(c1+(c2-c1)*(c0&0x7ff)/0x800));
#else
	    l += (Vol0*c1);
#endif
	    SCC[n].Counter0 = c0 ;
	}

	if (Vol1 && Incr1) {
	    c0 = (SCC[n].Counter1 + Incr1) & 0xffff ;
	    c1 = (signed char)SCC[n].Waves[0x20 + (c0>>11)] ;
#ifdef SCC_INTERPOLATE
	    c2 = (signed char)SCC[n].Waves[0x20 + ((c0+0x800)>>11)] ;
	    l += (Vol1*(c1+(c2-c1)*(c0&0x7ff)/0x800));
#else
	    l += (Vol1*c1);
#endif
	    SCC[n].Counter1 = c0 ;
	}

	if (Vol2 && Incr2) {
	    c0 = (SCC[n].Counter2 + Incr2) & 0xffff ;
	    c1 = (signed char)SCC[n].Waves[0x40 + (c0>>11)] ;
#ifdef SCC_INTERPOLATE
	    c2 = (signed char)SCC[n].Waves[0x40 + ((c0+0x800)>>11)] ;
	    l += (Vol2*(c1+(c2-c1)*(c0&0x7ff)/0x800));
#else
	    l += (Vol2*c1);
#endif
	    SCC[n].Counter2 = c0 ;
	}

	if (Vol3 && Incr3) {
	    c0 = (SCC[n].Counter3 + Incr3) & 0xffff ;
	    c1 = (signed char)SCC[n].Waves[0x60 + (c0>>11)] ;
#ifdef SCC_INTERPOLATE
	    c2 = (signed char)SCC[n].Waves[0x60 + ((c0+0x800)>>11)] ;
	    l += (Vol3*(c1+(c2-c1)*(c0&0x7ff)/0x800));
#else
	    l += (Vol3*c1);
#endif
	    SCC[n].Counter3 = c0 ;
	}

	if (Vol4 && Incr4) {
	    c0 = (SCC[n].Counter4 + Incr4) & 0xffff ;
	    c1 = (signed char)SCC[n].Waves[0x80 + (c0>>11)] ;
#ifdef SCC_INTERPOLATE
	    c2 = (signed char)SCC[n].Waves[0x80 + ((c0+0x800)>>11)] ;
	    l += (Vol4*(c1+(c2-c1)*(c0&0x7ff)/0x800));
#else
	    l += (Vol4*c1);
#endif
	    SCC[n].Counter4 = c0 ;
	}
	*buf++ = DATACONV (l);
	}
}

void SCCSetClock (int chip, int clock) {
    SCC[chip].UpdateStep = 512.0 * (double)clock /
	((double)SCC[chip].SampleRate / 4.0);
}

static int SCCInit(int chip, int clock,int volume,int sample_rate,int sample_bits) 
{
    memset (&SCC[chip],0,sizeof (struct KONSCC) );
    SCC[chip].SampleRate = sample_rate;

    //sprintf (name, "Konami SCC/SCC+ #%d",chip);

    /*SCC[chip].Channel = stream_init (name,volume,sample_rate,sample_bits,chip,
	(sample_bits == 16) ? SCCUpdate16 : SCCUpdate8);*/

	//SCC[chip].Channel = stream_init (name,volume,sample_rate,chip,SCCUpdate16);

    //if (SCC[chip].Channel == -1) return 1;

    SCCSetClock (chip, clock);
    SCCResetChip (chip);

    return 0;
}

int SCC_sh_start (int sample_freq) {
    if (SCCInit (1,3579545, 20,sample_freq, 16) )
		  return 1;
		return 0;	
}
